use crate::value::{decoder, encoder, invert_map, NestedValue};
use decoder::{Decoder, DecodePlaceholderMap};
use encoder::{Encoder, EncodePlaceholderMap};
use std::io::{Read, Write};

pub struct Codec<N: NestedValue> {
    pub decode_placeholders: Option<DecodePlaceholderMap<N>>,
    pub encode_placeholders: Option<EncodePlaceholderMap<N>>,
}

impl<N: NestedValue> Codec<N> {
    pub fn new(decode_placeholders: DecodePlaceholderMap<N>) -> Self {
        let encode_placeholders = Some(invert_map(&decode_placeholders));
        Codec { decode_placeholders: Some(decode_placeholders), encode_placeholders }
    }

    pub fn without_placeholders() -> Self {
        Codec { decode_placeholders: None, encode_placeholders: None }
    }

    pub fn decoder<'a, 'r, R: Read>(&'a self, read: &'r mut R) -> Decoder<'r, 'a, R, N> {
        Decoder::new(read, self.decode_placeholders.as_ref())
    }

    pub fn encoder<'a, 'w, W: Write>(&'a self, write: &'w mut W) -> Encoder<'w, 'a, W, N> {
        Encoder::new(write, self.encode_placeholders.as_ref())
    }

    pub fn decode<'r, R: Read>(&self, read: &'r mut R) -> decoder::Result<N> {
        self.decoder(read).next()
    }

    pub fn encode_bytes(&self, v: &N) -> std::io::Result<Vec<u8>> {
        let mut buf: Vec<u8> = Vec::new();
        self.encoder(&mut buf).write(v)?;
        Ok(buf)
    }
}
